package ch.epfl.collections;

import static org.junit.Assert.*;

import java.util.Iterator;

import org.junit.Test;

public abstract class ListTest {
    abstract List<String> newList();

    @Test
    public void testAdd() {
        List<String> l = newList();
        l.add("hello");
        assertEquals("hello", l.get(0));
        assertEquals(1, l.size());
    }

    @Test(expected = IndexOutOfBoundsException.class)
    public void testGetEmpty() {
        List<String> l = newList();
        l.get(0);
    }

    @Test
    public void testGet() {
        List<String> l = newList();
        l.add("hello");
        l.add("world");
        assertEquals("hello", l.get(0));
        assertEquals("world", l.get(1));
    }

    @Test
    public void testSet() {
        List<String> l = newList();
        l.add("hello");
        l.set(0, "world");
        assertEquals("world", l.get(0));
    }

    @Test
    public void testSize() {
        List<String> l = newList();
        assertEquals(0, l.size());
        l.add("hello");
        assertEquals(1, l.size());
        l.remove(0);
        assertEquals(0, l.size());
    }

    @Test
    public void testIsEmpty() {
        List<String> l = newList();
        assertTrue(l.isEmpty());
        l.add("hello");
        assertFalse(l.isEmpty());
    }

    @Test
    public void testRemove() {
        List<String> l = newList();
        l.add("one");
        l.add("two");
        l.add("three");
        l.remove(1);
        assertEquals(2, l.size());
        assertEquals("one", l.get(0));
        assertEquals("three", l.get(1));
    }

    @Test
    public void testRemoveHead(){
        List<String> l = newList();
        l.add("one");
        l.add("two");
        l.add("three");
        l.remove(0);
        assertEquals(2, l.size());
        assertEquals("two", l.get(0));
        assertEquals("three", l.get(1));
    }

    @Test
    public void testIterator() {
        List<String> l = newList();
        String[] a = new String[] { "one", "two", "three", "four" };
        for (String s: a)
            l.add(s);
        Iterator<String> it = l.iterator();
        int i = 0;
        while (it.hasNext()) {
            String e = it.next();
            assertEquals(a[i++], e);
        }
        assertEquals(i, a.length);
    }

    @Test
    public void testIteratorForEach() {
        List<String> l = newList();
        String[] a = new String[] { "one", "two", "three", "four" };
        for (String s: a)
            l.add(s);
        int i = 0;
        for (String e: l)
            assertEquals(a[i++], e);
        assertEquals(i, a.length);
    }


    @Test
    public void testIteratorRemove() {
        List<String> l = newList();
        for (int i = 0; i <= 9; ++i)
            l.add(String.valueOf(i));
        Iterator<String> it = l.iterator();
        int i = 0;
        while (it.hasNext()) {
            String next = it.next();
            assertEquals(String.valueOf(i++), next);
            it.remove();
        }
        assertTrue(l.isEmpty());
    }

    @Test
    public void testIteratorRemove2() {
        List<String> l = newList();
        for (int i = 0; i < 10; ++i)
            l.add(String.valueOf(i));
        Iterator<String> it = l.iterator();
        while (it.hasNext()) {
            it.next();
            it.next();
            it.remove();
        }
        assertEquals(5, l.size());
        int i = 0;
        for (String s: l)
            assertEquals(String.valueOf((i++)*2), s);
    }

    @Test(expected = IllegalStateException.class)
    public void testIteratorRemoveWithoutNext() {
        newList().iterator().remove();
    }

    @Test(expected = IllegalStateException.class)
    public void testIteratorRemoveTwice() {
        List<String> l = newList();
        l.add("hello");
        l.add("world");
        Iterator<String> i = l.iterator();
        i.next();
        i.remove();
        i.remove();
    }
}
